@@ -16,7 +16,7 @@ if __name__ == '__main__':
         assert False, (r"parser.add_argument('-f', '--file', help='The node name is?')")
 
     if hasattr(args, 'port'):
-        port = args.port
+        port = int(args.port)
     else:
         port = find_free_port()
         print('没有用--port指定端口，自动查找到可用端口:', port)
